"""

    ConvexRegistry

"""
import logging

from convex_api.exceptions import ConvexAPIError
from convex_api.utils import to_address


logger = logging.getLogger(__name__)

QUERY_ACCOUNT_ADDRESS = 9


class ConvexRegistry:

    def __init__(self, convex):
        self._convex = convex
        self._address = None
        self._items = {}

    def is_registered(self, name):
        return self.item(name) is not None

    def item(self, name):
        if name not in self._items:
            result = self._convex.query(f'(get cns-database (symbol "{name}"))', self.address)
            logger.debug(f'cns-database: {result}')
            self._items[name] = None
            if result and 'value' in result and isinstance(result['value'], (list, tuple)):
                self._items[name] = result['value']
        return self._items[name]

    def clear(self):
        self._items = {}
        self._address = None

    def register(self, name, contract_address, account):
        result = self._convex.send(f'(call {self.address} (register {{:name "{name}"}}))', account)
        logger.debug(f'register result: {result}')
        if result and 'value' in result:
            try:
                result = self._convex.send(f'(call {self.address} (cns-update "{name}" {contract_address}))', account)
                logger.debug(f'cns-update result: {result}')
                if result and 'value' in result:
                    items = result['value']
                    if name in items:
                        item = items[name]
                        self._items[name] = item
                        return item
            except ConvexAPIError as e:
                logger.debug(f'convex error: {e}')

    def resolve_owner(self, name):
        item = self.item(name)
        if item:
            return item[1]

    def resolve_address(self, name):
        item = self.item(name)
        if item:
            return item[0]

    @property
    def address(self):
        if self._address is None:
            result = self._convex.query('(address *registry*)', QUERY_ACCOUNT_ADDRESS)
            self._registry_address = to_address(result['value'])
            logger.debug(f'registry_address: {self._registry_address}')
        return self._registry_address



"""

    Exceptions for starfish-py

"""


class StarfishConnectionError(Exception):
    """ Raised when starfish cannot connect to the network, or remote agent """


class StarfishPurchaseError(Exception):
    """ Raised when a purchase events have failed to complete """


class StarfishAssetNotFound(Exception):
    """ Raised when an asset is not found """


class StarfishAssetInvalid(Exception):
    """ Raised when a downloaded asset is not valid or has been changed """


class StarfishInsufficientFunds(Exception):
    """ Raised when the account has insufficient funds to send token or ether """


class StarfishRemoteAgentInvalidAccess(Exception):
    """ Raised when a access token can not be generated by the remote agent """


class StarfishContractError(Exception):
    """ Raised when a contract cannot be loaded or accesed """



"""
    Inovke Asset
"""
import json
from typing import (
    Any,
    Generic
)

from starfish.asset.asset_base import AssetBase
from starfish.types import TOperationAsset


class OperationAsset(AssetBase, Generic[TOperationAsset]):

    """

    Operation asset can be used to perform a invokable service

    :param metadata: Optional dictionary metadata to provide for the asset
        if non used then the class will generate a default metadata based on the data provided
    :type metadata: None or dict
    :param did: Optional did of the asset if it's registered
    :type did: None or str

    """
    def __init__(self, metadata_text: str) -> None:
        AssetBase.__init__(self, metadata_text)

    def create(name: str, metadata: Any = None) -> TOperationAsset:
        """

        Create a new OperationAsset.

        :param str name: Name of the asset
        :param dict metadata: Optional metadata to add to the assets metadata
        :param str did: Option DID to assign to this asset

        """
        metadata = AssetBase.generateMetadata(name, 'operation', metadata)
        return OperationAsset(json.dumps(metadata))

    def is_mode(self, mode_type: str) -> bool:
        """

        Check to see if this operation supports the mode provided.

        :param str mode_type: Mode type to check to see if this operation supports
        :return: Return True if this mode is supported
        :type: boolean

        """
        try:
            return mode_type in self.metadata['operation']['modes']
        except ValueError:
            raise ValueError('Metadata does not contain operation->modes structure')




import json
import pytest
import secrets
import tempfile
import os

from web3 import Web3

from starfish.network.ethereum.ethereum_account import EthereumAccount


def test_init(config):
    account = EthereumAccount.import_from_text(config.ethereum.accounts[0].key_data, config.ethereum.accounts[0].password)
    assert(account)
    assert(account.address)

def test_is_address_equal(config):
    account = EthereumAccount.import_from_text(config.ethereum.accounts[0].key_data, config.ethereum.accounts[0].password)
    assert(account.is_address_equal(config.ethereum.accounts[0].address))

def test_is_valid(config):
    account = EthereumAccount.import_from_text(config.ethereum.accounts[0].key_data, config.ethereum.accounts[0].password)
    assert(account.is_valid)

def test_is_password(config):
    account = EthereumAccount.import_from_text(config.ethereum.accounts[0].key_data, config.ethereum.accounts[0].password)
    assert(account.password == config.ethereum.accounts[0].password)

def test_is_key_data(config):
    account = EthereumAccount.import_from_text(config.ethereum.accounts[0].key_data, config.ethereum.accounts[0].password)
    assert(account.key_data == config.ethereum.accounts[0].key_data)

def test_address(config):
    account = EthereumAccount.import_from_text(config.ethereum.accounts[0].key_data, config.ethereum.accounts[0].password)
    assert(account.address.lower() == config.ethereum.accounts[0].address.lower())

def test_as_checksum_address(config):
    account = EthereumAccount.import_from_text(config.ethereum.accounts[0].key_data, config.ethereum.accounts[0].password)
    assert(account.as_checksum_address == Web3.toChecksumAddress(config.ethereum.accounts[0].address))

def test_password(config):
    account = EthereumAccount.import_from_text(config.ethereum.accounts[0].key_data, config.ethereum.accounts[0].password)
    assert(account.password == config.ethereum.accounts[0].password)

def test_account_key_file():
    test_password = secrets.token_bytes(32)
    account = EthereumAccount.create(test_password)

    assert(account.password == test_password)
    assert(json.dumps(account.key_data))

    test_filename = '/tmp/key_file.json'
    account.export_to_file(test_filename)
    assert(os.path.exists(test_filename))
    account_import = EthereumAccount.import_from_file(test_filename, test_password)
    os.unlink(test_filename)

    assert(account_import.is_address_equal(account.address))
    account.export_to_file(test_filename)
    assert(os.path.exists(test_filename))







from starfish.network.ethereum.ethereum_account import EthereumAccount                                   # noqa: F401
from starfish.network.ethereum.ethereum_network import EthereumNetwork                                   # noqa: F401



"""

    Dex Token Contract

"""
from starfish.network.ethereum.ethereum_account import EthereumAccount
from starfish.types import AccountAddress

from .contract_base import ContractBase


CONTRACT_NAME = 'DatacraftToken'


class DatacraftTokenContract(ContractBase):
    """Class representing the Dex Token contract."""

    def __init__(self) -> None:
        ContractBase.__init__(self, CONTRACT_NAME)

    def get_balance(self, account_address: AccountAddress) -> float:
        address = self.get_account_address(account_address)
        amount_wei = self.call('balanceOf', address)
        return self.to_ether(amount_wei)

    def approve_transfer(self, account: EthereumAccount, to_account_address: AccountAddress, amount: float) -> bool:
        to_address = self.get_account_address(to_account_address)
        amount_wei = self.to_wei(amount)
        return self.call('approve', (to_address, amount_wei), account)

    def transfer(self, account: EthereumAccount, to_account_address: AccountAddress, amount: float) -> bool:
        to_address = self.get_account_address(to_account_address)
        amount_wei = self.to_wei(amount)
        return self.call('transfer', (to_address, amount_wei), account)

    def total_supply(self) -> float:
        amount_wei = self.call('totalSupply')
        return self.to_ether(amount_wei)



"""


Contract Base

"""

from convex_api import ConvexAPI

from starfish.network.convex.contract.convex_registry import ConvexRegistry


class ContractBase:
    def __init__(self, convex: ConvexAPI, name: str):
        self._convex = convex
        self._name = name
        self._registry = ConvexRegistry(convex)
        self._version = None
        self._address = None
        self._owner_address = None

    def send(self, transaction, account):
        if not self.address:
            raise ValueError(f'No contract address found for {self._name}')
        return self._convex.send(f'(call {self.address} {transaction})', account)

    def query(self, transaction, account_address=None):
        if account_address is None:
            account_address = self.address
        if not self.address:
            raise ValueError(f'No contract address found for {self._name}')
        return self._convex.query(f'(call {self.address} {transaction})', account_address)

    @property
    def deploy_version(self):
        if self.address:
            result = self.query('(version)')
            if result and 'value' in result:
                return result['value']

    @property
    def is_registered(self):
        return self.address is not None

    @property
    def address(self):
        if self._address is None:
            self._address = self._registry.resolve_address(self._name)
        return self._address

    @property
    def owner_address(self):
        if self._owner_address is None:
            self._owner_address = self._registry.resolve_owner(self._name)
        return self._owner_address

    @property
    def name(self):
        return self._name

    @property
    def version(self):
        return self._version



"""


    AgentAccess


"""
import hashlib
import json
import logging


from starfish.agent import RemoteAgent
from starfish.exceptions import StarfishConnectionError
from starfish.network.ddo import DDO
from starfish.network.did import (
    did_to_id,
    is_did
)

logger = logging.getLogger(__name__)


class AgentAccess:

    def __init__(
        self,
        name,
        ddo_text=None,
        did=None,
        url=None,
        authentication=None,
        username=None,
        password=None,
        token=None,
        http_client=None
    ):
        """
        Create an agent access record. This record tries to obtain all the nesseray information needed to resolve
        an agent.

        :param str name: Name of the agent, this can be used to get the agent information.
        :param str ddo_text: DDO text of the agent if available, if not this class will request for it later.
        :param str did: DID of the agent, similar to ddo_text, the did wil requested later if not available.
        :param str url: URL of the agent.
        :param dict authentication: authentication dict to access the agent
        :param str username: username to access the agent, this can also be in the authentication.
        :param str password: password to access the agent, this can also be in the authentication.
        :param str token: token to access the agent, this can also be in the authentication.

        """
        ddo = None
        if ddo_text:
            ddo = DDO.import_from_text(ddo_text)
            if did is None:
                did = ddo.did

        if authentication is None and (username or token):
            authentication = {}
            if username:
                authentication['username'] = username

            if password:
                authentication['password'] = password

            if token:
                authentication['token'] = token

        self._name = name
        self._url = url
        self._ddo = ddo
        self._did = did
        self._authentication = authentication
        self._agent_cache = {}
        self._http_client = http_client

    @staticmethod
    def resolve_agent_url(url, authentication=None, http_client=None):
        """
        Resolve an agent based on it's url.

        :param str url: URL of the agent
        :param dict authentication: Authentication of the agent.
        :param object http_client: HTTP client to use to access the agent api

        :returns: DDO object of the remote agent
        """
        logger.debug(f'resolving remote agent did from {url}')
        try:
            ddo_text = RemoteAgent.resolve_url(url, authentication=authentication, http_client=http_client)
            if ddo_text:
                return DDO.import_from_text(ddo_text)
        # ignore connetion errors to remote agents
        except StarfishConnectionError:
            pass

    @staticmethod
    def resolve_agent_did(did, network, authentication=None, http_client=None):
        """
        Resolve an agent based on it's DID.

        :param str did: DID of the agent
        :param dict authentication: Authentication of the agent.
        :param object http_client: HTTP client to use to access the agent api

        :returns: DDO object of the remote agent
        """
        return network.resolve_did(did)

    def resolve_url(self):
        """
        Resolve the remote agent using it's URL.

        :param object http_client: HTTP Client to use to access the remote agent api.

        :returns: DDO of the remote agent

        """
        if self._url:
            ddo = AgentAccess.resolve_agent_url(self._url, self._authentication, self._http_client)
            if ddo:
                self._ddo = ddo
                self._did = ddo.did
                return ddo

    def resolve_did(self, network):
        if self._did:
            ddo = AgentAccess.resolve_agent_did(self._did, network)
            if ddo:
                self._ddo = ddo
                self._did = ddo.did
                return ddo

    def is_match(self, name_did_url):
        """
        Is this AgentAccess object matches the name, DID or URL.

        :param str name_did_url: Name, DID or URL to match

        :returns: True if match

        """
        if self._name == name_did_url:
            return True
        if self._url == name_did_url:
            return True
        if self._did and is_did(self._did) and is_did(name_did_url):
            if did_to_id(self._did) == did_to_id(name_did_url):
                return True
        return False

    def load_agent(self, authentication=None, use_cache=True):
        """
        Load this agent and return a starfish RemoteAgent object.

        :param dict authentication: Authentication dict to use to load this agent.
            If None use the authentication in the access object

        :param bool use_cache: If True then use the agent cache to store future requests for remote agents
        :param object http_client: HTTP Client to use to access the agent.

        :returns: Remote Agent object that has been loaded

        """
        agent = None
        if authentication is None:
            authentication = self._authentication
        if use_cache:
            cache_key = self.calc_cache_key(authentication)
            if cache_key not in self._agent_cache:
                # same method but with no cache
                self._agent_cache[cache_key] = self.load_agent(authentication, use_cache=False)
            agent = self._agent_cache[cache_key]
        else:
            if self._ddo is None:
                self.resolve_url(self._http_client)
            logger.debug(f'loading remote agent {self._name}: {self._did}')
            agent = RemoteAgent(self._ddo, authentication=authentication, http_client=self._http_client)

        return agent

    def calc_cache_key(self, authentication=None):
        key_text = [
            self._name
        ]
        if authentication:
            key_text.append(json.dumps(authentication))
        md5hash = hashlib.md5('\n'.join(key_text).encode())
        return md5hash.hexdigest()

    def clear_cache(self):
        """
        Clears out the agent cache

        """
        self._agent_cache = {}

    @property
    def name(self):
        return self._name

    @property
    def url(self):
        return self._url

    @property
    def ddo(self):
        return self._ddo

    @property
    def did(self):
        return self._did

    @property
    def authentication(self):
        return self._authentication

    @property
    def http_client(self):
        return self._http_client

    @http_client.setter
    def http_client(self, value):
        self._http_client = value

    def __str__(self):
        return f'{self._name} {self._url} {self._did}'



"""


    Contract Manager

    loads contracts


"""

import gzip
import importlib
import inspect
import json
import logging
import os
import time
from typing import (
    Any,
    Generic
)
from urllib.parse import urljoin
import requests

from starfish.network.ethereum.contract.contract_base import ContractBase
from starfish.network.ethereum.ethereum_network import EthereumNetwork
from starfish.types import (
    TContractBase,
    TContractManager,
    TNetworkBase
)

logger = logging.getLogger(__name__)


ARTIFACT_DATA_FILENAME = 'artifacts.json.gz'
LOCAL_ARTIFACT_PACKAGE_SERVER = 'http://localhost:8550'


class ContractManager(Generic[TContractManager]):
    """
    Setup the contract manager to load and get contracts

    :param network: Network connection to the block chain
    :param str default_package_name: default package name for the contract module

    """

    def __init__(self, network: EthereumNetwork, default_package_name: str, artifacts_path: str) -> None:
        self._network = network
        self._default_package_name = default_package_name
        self._artifacts_path = artifacts_path
        self._artifact_items = ContractManager.load_artifacts_package(ARTIFACT_DATA_FILENAME)
        if self._artifact_items is None:
            self._artifact_items = []

    def load(self, name: str, artifact_filename: str = None, has_artifact: bool = None, package_name: str = None) -> TContractBase:
        """

        Load a contract using it's name, and network name

        :param str name: Name of the contract to load
        :param str artifact_filename: Optional filename of the artifact file
        :param bool has_artifact: Defaults to True, if false then just load the contract without an abi & address
        :param str package_name: Defaults to default_package_name, the package for the contract module

        :return: Return a contract object, else return None

        """
        if package_name is None:
            package_name = self._default_package_name
        if package_name is None:
            raise ValueError('You need to provide a package name for the contract classes')

        if has_artifact is None:
            has_artifact = True

        package_module = importlib.import_module(package_name)
        for item in inspect.getmembers(package_module, inspect.ismodule):
            class_def = ContractManager._find_class_in_module(name, item[1])
            if class_def:
                return self.create_contract_object(
                    name,
                    class_def,
                    artifact_filename=artifact_filename,
                    has_artifact=has_artifact
                )
        return None

    def load_local_artifacts_package(self, url: str = None, timeoutSeconds: int = 20) -> None:
        if not url:
            url = LOCAL_ARTIFACT_PACKAGE_SERVER

        timeout = time.time() + timeoutSeconds
        info = self._request_local_artifacts_package(url)
        while not info and timeout < time.time():
            time.sleep(1)
            info = self._request_local_artifacts_package(url)
        if info and 'artifacts' in info:
            self._artifact_items = info['artifacts']

    def create_contract_object(self, name: str, class_def: Any, artifact_filename: str = None, has_artifact: bool = None) -> Any:
        """

        Create a new contract object from a class definition

        :param str name: Name of the contract
        :param class class_def: Class definition of the contract
        :param str artifact_filename: Optional filename of the artifact file
        :param bool has_artifact: Defaults to True, if false then just load the contract without an abi & address

        :return: Return a contract object, else return None

        """
        if has_artifact is None:
            has_artifact = True

        contract_object = class_def()
        contract_name = contract_object.name

        # only load in an artifact file if it is needed by the contract
        if has_artifact:
            contract_data = self.get_contract_data(contract_name, artifact_filename)
            if contract_data:
                contract_object.load(self._network.web3, abi=contract_data['abi'], address=contract_data['address'])
            else:
                raise FileNotFoundError(f'Cannot find artifact data for contract {contract_name}.{self._network.name}')
        else:
            # load in an dummy contract with no abi or address
            contract_object.load(self._network.web3)

        return contract_object

    def get_contract_data(self, name: str, artifact_filename: str = None) -> Any:
        """

        Get the contract article data from a file or from the library.

        :param str name: Name of the contract
        :param str artifact_filename: Optional filename of the artifact file

        :return: Return a contract article data or None

        """
        data = None

        # setup the default artifact filename
        if artifact_filename is None:
            artifact_filename = f'{name}.{self._network.network_id}.json'

        # first look for user defined folder for any artifact file
        artifacts_path_list = [self._artifacts_path, ContractManager.data_path()]
        artifact_filename_path = ContractManager.find_artifact_filename(artifact_filename, artifacts_path_list)
        if artifact_filename_path:
            logger.debug(f'loading contract artifact file at {artifact_filename_path}')
            data = ContractManager.load_artifact_file(artifact_filename_path)
        else:
            # now try to load from the artifact library
            if (str(self._network.network_id) in self._artifact_items and
                    name in self._artifact_items[str(self._network.network_id)]):
                logger.debug(f'found contract in library {name}.{self._network.network_id}')
                data = self._artifact_items[str(self._network.network_id)][name]

        return data

    def set_contract_data(self, name: str, data: Any) -> None:
        """

        Set the contract article data for given contract.

        :param str name: Name of the contract to set
        :param dict data: Article data for the contract

        """
        if self._network.network_id not in self._artifact_items:
            self._artifact_items[self._network.network_id] = {}
        self._artifact_items[self._network.network_id][name] = data

    def _request_local_artifacts_package(self, url: str) -> Any:
        """
        Request the artifacts package from the local package serever.
        This only works for local testing with a local private network.

        Usually the artifacts are stored in a single gzipped json file for all
        of the public and test block chain networks.
        :param str url: URL of the package server to request
        :returns: a dict on artifacts
        """
        data = None
        url = urljoin(f'{url}/', 'artifacts')
        logger.debug(f'requesting artifacts at url {url}')
        response = requests.get(url)
        if response and response.status_code == requests.codes.ok:
            data = response.json()
        return data

    @property
    def network(self) -> TNetworkBase:
        return self._network

    @property
    def artifacts_path(self) -> str:
        return self._artifacts_path

    @staticmethod
    def data_path() -> str:
        return os.path.join(os.path.dirname(__file__), 'data')

    @staticmethod
    def _find_class_in_module(class_name: str, contract_module: str) -> Any:
        for name, obj in inspect.getmembers(contract_module, inspect.isclass):
            if issubclass(obj, ContractBase) \
               and name != 'ContractBase' \
               and name == class_name:
                return obj
        return None

    @staticmethod
    def find_artifact_filename(filename: str, path_list: str) -> str:
        found_file = None
        for path in path_list:
            test_file = os.path.join(path, filename)
            if os.path.exists(test_file):
                found_file = test_file
                break

        return found_file

    @staticmethod
    def load_artifact_file(filename: str) -> Any:
        if filename and os.path.exists(filename):
            with open(filename, 'r') as fp:
                return json.load(fp)

    @staticmethod
    def load_artifacts_package(filename: str) -> Any:
        data = None
        artifact_libray_file = os.path.join(ContractManager.data_path(), filename)
        if os.path.exists(artifact_libray_file):
            with gzip.open(artifact_libray_file, 'rt') as fp:
                data = json.load(fp)
        return data





from starfish.network.convex.convex_account import ConvexAccount                             # noqa: F401
from starfish.network.convex.convex_network import ConvexNetwork                             # noqa: F401



"""

    Test DDO


"""

import json
import pytest

from starfish.network.ddo import DDO
from starfish.network.did import did_generate_random


def test_ddo_create_from_service_list():
    test_url = 'http://localhost'
    ddo = DDO.create(test_url, service_list=['meta', 'trust'])
    assert(ddo)
    #ddo_text = json.dumps(json.loads(ddo.as_text), sort_keys=True, indent=4)
    #print(ddo_text)
    assert('meta' in ddo.service)
    assert(ddo.service['meta']['serviceEndpoint'] == 'http://localhost/api/v1/meta')

    with pytest.raises(TypeError, match=r'Service list must be a list or tuple of service names'):
        ddo = DDO.create(test_url, service_list={
            'bad_name':'meta',
            'invalid': 'trust',
        })

    with pytest.raises(TypeError, match=r'Service list must be a list of type string names'):
        ddo = DDO.create(test_url, service_list=[
            {'type':'meta'},
            {'type': 'trust'},
        ])

def test_ddo_add_services():
    ddo = DDO()
    ddo.add_service('meta', endpoint='http://localhost/api/v98/meta')
    ddo.add_service('trust', url='http://localhost')
    ddo.add_service('DID.Meta.v1', endpoint='http://localhost/api/v1/meta_did')


    assert('meta' in ddo.service)
    assert(ddo.service['meta']['serviceEndpoint'] == 'http://localhost/api/v98/meta')

    assert('trust' in ddo.service)
    assert(ddo.service['trust']['serviceEndpoint'] == 'http://localhost/api/v1/trust')

    assert('DID.Meta.v1' in ddo.service)
    assert(ddo.service['DID.Meta.v1']['serviceEndpoint'] == 'http://localhost/api/v1/meta_did')

    with pytest.raises(ValueError, match=r'Invalid service name'):
        ddo.add_service('test')

    with pytest.raises(ValueError, match=r'No endpoint or url provided to add a service to the ddo'):
        ddo.add_service('auth')



def test_ddo_create_from_service_list_version():
    test_url = 'http://localhost'
    ddo = DDO.create(test_url, service_list=['meta', 'trust'], version='v99')
    assert(ddo)
    #ddo_text = json.dumps(json.loads(ddo.as_text), sort_keys=True, indent=4)
    #print(ddo_text)
    assert('meta' in ddo.service)
    assert(ddo.service['meta']['serviceEndpoint'] == 'http://localhost/api/v99/meta')


def test_ddo_create_all_services():
    test_url = 'http://localhost'
    ddo = DDO.create(test_url)
    assert(ddo)
    assert(len(ddo.service_list) > 5)


def test_ddo_basic_init():
    did = did_generate_random()
    ddo = DDO(did)
    assert(ddo)
    assert(ddo.did == did)


def test_ddo_import_from_text():
    test_url = 'http://localhost'
    ddo = DDO.create(test_url, service_list=['meta', 'trust'])
    assert(ddo)
    ddo_text = json.dumps(json.loads(ddo.as_text), sort_keys=True, indent=4)
    import_ddo = DDO.import_from_text(ddo_text)
    assert(import_ddo)
    assert(import_ddo.as_text == ddo.as_text)

def test_ddo_is_supported_service():
    assert(DDO.is_supported_service('meta'))
    assert(not DDO.is_supported_service('bad-service-name'))


def test_ddo_get_did_from_ddo():
    did = did_generate_random()
    test_url = 'http://localhost'
    ddo = DDO.create(test_url, service_list=['meta', 'trust'], did=did)
    assert(ddo)
    assert(ddo.did == did)
    assert(DDO.get_did_from_ddo(ddo) == did)
    assert(DDO.get_did_from_ddo(ddo.as_text) == did)

def test_ddo_checksum():
    test_url = 'http://localhost'
    ddo = DDO.create(test_url, service_list=['meta', 'trust'])
    assert(ddo.checksum == '5da48d1d8b9ad1fdb27400e336dc007dc90a86957dfe1cc179566118d79d9166')



"""

    Command Account Get ..

"""
from typing import Any

from .account_get_ether_command import AccountGetEtherCommand
from .account_get_token_command import AccountGetTokenCommand
from .command_base import CommandBase
from .help_command import HelpCommand

DEFAULT_AMOUNT = 10


class AccountGetCommand(CommandBase):

    def __init__(self, sub_parser: Any = None) -> None:
        self._command_list = []
        super().__init__('get', sub_parser)

    def create_parser(self, sub_parser: Any) -> Any:

        parser = sub_parser.add_parser(
            self._name,
            description='Get ether or tokens for an account address',
            help='Get some token or ether from a faucet or a tokens on a test network'

        )
        account_get_parser = parser.add_subparsers(
            title='get command',
            description='get command',
            help='Get command',
            dest='get_command'
        )

        self._command_list = [
            AccountGetEtherCommand(account_get_parser),
            AccountGetTokenCommand(account_get_parser),
            HelpCommand(account_get_parser, self)
        ]
        return account_get_parser

    def execute(self, args: Any, output: Any) -> Any:
        return self.process_sub_command(args, output, args.get_command)



"""
    test_asset_base

"""

import secrets
import logging
import json
import pytest

from starfish.asset import AssetBase
from starfish.network.did import (
    did_to_id,
    id_to_did,
    decode_to_asset_id,
    did_generate_random
)

ASSET_METADATA = {
    'name': 'Asset',
    'type': 'asset',
}

TEST_DID = did_generate_random()

def test_init(metadata):
    asset = AssetBase(json.dumps(ASSET_METADATA))
    assert(asset)
    assert(isinstance(asset, AssetBase))

def test_metadata():
    asset = AssetBase(json.dumps(ASSET_METADATA))
    assert(asset)
    assert(asset.metadata == ASSET_METADATA)

def test_data():
    asset = AssetBase(json.dumps(ASSET_METADATA))
    assert(asset)
    assert(asset.metadata == ASSET_METADATA)
    asset.set_did(TEST_DID)
    assert(asset.did == TEST_DID)


def test_is_asset_type():
    asset = AssetBase(json.dumps(ASSET_METADATA))
    assert(asset)
    assert(asset.is_asset_type('asset'))
    assert(not asset.is_asset_type('bad asset type'))

def test_asset_id():
    asset = AssetBase(json.dumps(ASSET_METADATA))
    assert(asset)
    asset.set_did(TEST_DID)
    with pytest.raises(ValueError):
        asset_id = decode_to_asset_id(asset.did)

    test_did = did_generate_random()
    asset_id = secrets.token_hex(32)
    asset_did = f'{test_did}/{asset_id}'

    asset = AssetBase(json.dumps(ASSET_METADATA))
    assert(asset)
    asset.set_did(asset.did)
    #assert(asset.asset_id == asset_id)

def test_asset_add_provenance_publish():
    agent_did = did_generate_random()
    asset = AssetBase(json.dumps(ASSET_METADATA))
    assert(asset)
    asset.add_provenance_publish(agent_did)
    assert(asset.metadata['provenance'])


def test_asset_add_provenance_invoke():
    agent_did = did_generate_random()
    job_id = secrets.token_hex(32)
    inputs_text = json.dumps(
        {
            'asset_list': 'json'
        }
    )
    asset = AssetBase(json.dumps(ASSET_METADATA))
    assert(asset)
    asset.add_provenance_invoke(agent_did, job_id, None, inputs_text)
    assert(asset.metadata['provenance'])



"""


    Local node Utils to get local node artifact files, from a docker container


"""

import json
import logging
import os
import re
import tarfile
import tempfile
from typing import Any
import docker


logger = logging.getLogger(__name__)


def get_local_contract_files(docker_name: str, from_folder: str) -> Any:
    items = []
    client = docker.from_env()
    container_list = client.containers.list(filters={'name': 'keeper-contracts'})
    if container_list:
        container = container_list[0]
        logger.debug(f'found container id: {container.id} name: {container.name} state: {container.status}')
        items = export_docker_contract_items(container, from_folder)
    return items


def export_docker_contract_items(container: Any, from_folder: str) -> Any:
    items = {}
    data, stat = container.get_archive(from_folder)
    if stat:
        with tempfile.TemporaryFile() as file_handle:
            for chunk in data:
                file_handle.write(chunk)
            file_handle.seek(0, 0)
            tar = tarfile.open(fileobj=file_handle)
            if tar:
                tar_item = tar.next()
                while tar_item:
                    if tar_item.isfile() and re.search(r'\.json$', tar_item.name):
                        filename = os.path.basename(tar_item.name)
                        # logger.debug(f'reading contract file {filename}, {tar_item.size}')
                        with tar.extractfile(tar_item) as fp:
                            contract_data = json.load(fp)
                            items[filename] = contract_data
                    tar_item = tar.next()
    return items



"""

    Contactt Base

"""
from typing import (
    Any,
    Generic
)

from starfish.network.ethereum.ethereum_account import EthereumAccount
from starfish.types import (
    AccountAddress,
    TContractBase
)

nonce_list = {}

GAS_MINIMUM = 200000


class ContractBase(Generic[TContractBase]):

    def __init__(self, name: str) -> None:
        self._name = name
        self._web3 = None
        self._abi = None
        self._address = None
        self._contract = None

    def load(self, web3: Any, abi: Any = None, address: str = None) -> None:
        self._web3 = web3
        self._abi = abi
        self._address = address
        self._contract = None
        if abi and address:
            self._contract = self._web3.eth.contract(
                address=address,
                abi=abi
            )

    def call(self, function_name: str, parameters: Any, account: EthereumAccount = None, transact: Any = None) -> Any:
        if self._contract is None:
            raise ValueError('contract not loaded')

        if not isinstance(parameters, (list, tuple)):
            parameters = (parameters,)

        contract_function_call = self._contract.functions[function_name](*parameters)
        if account:
            result = self._call_as_transaction(contract_function_call, account, transact)
            if result is None:
                raise ValueError(f'unable to sign a transaciton for {self.name}:{function_name}')
        else:
            result = contract_function_call.call()
        return result

    def get_event(self, event_name: str, parameters: Any = None) -> Any:
        if self._contract is None:
            raise ValueError('contract not loaded')
        return self._contract.events[event_name]

    def create_event_filter(
            self, event_name: str,
            parameters: Any = None,
            from_block: int = 1,
            argument_filters: Any = None
            ) -> Any:
        event = self.get_event(event_name, parameters)
        if event:
            return event.createFilter(
                fromBlock=from_block,
                argument_filters=argument_filters
            )
        return None

    def _call_as_transaction(self, contract_function_call: Any, account: EthereumAccount, transact: Any = None) -> str:
        if transact is None:
            gas_transact = {
                'from': account.address
            }
            gas = contract_function_call.estimateGas(gas_transact)
            transact = {
                'from': account.address,
                'gas': gas,
                'nonce': self.get_nonce(account.address)
            }

        built_transaction = contract_function_call.buildTransaction(transact)
        transaction = {
            'from': account.address,
            'to': built_transaction['to'],
            'gas': built_transaction['gas'],
            'data': built_transaction['data'],
            'gasPrice':  self.get_gas_price(account.address),
            'nonce':  built_transaction['nonce'],
        }
        signed = account.sign_transaction(transaction, self._web3)
        tx_hash = None
        if signed:
            tx_hash = self._web3.eth.sendRawTransaction(signed.rawTransaction)

        return tx_hash

    def wait_for_receipt(self, tx_hash: str, timeout: int = 30) -> Any:
        self._web3.eth.waitForTransactionReceipt(tx_hash, timeout=timeout)
        return self._web3.eth.getTransactionReceipt(tx_hash)

    def get_nonce(self, address: str) -> int:
        global nonce_list
        nonce = self._web3.eth.getTransactionCount(address)
        if address not in nonce_list:
            nonce_list[address] = nonce
        else:
            if nonce <= nonce_list[address]:
                nonce = nonce_list[address] + 1
        return nonce

    def get_gas_price(self, address: str) -> int:
        block = self._web3.eth.get_block("latest")
        gas_price = int(self._web3.eth.gasPrice / 100)
        gas_price = min(block.gasLimit, gas_price)
        gas_price = max(GAS_MINIMUM, gas_price)
        return gas_price

    def unlockAccount(self, account: EthereumAccount) -> None:
        self._web3.personal.unlockAccount(account.address, account.password)

    def lockAccount(self, account: EthereumAccount) -> None:
        self._web3.personal.lockAccount(account.address, account.password)

    @property
    def name(self) -> str:
        return self._name

    @property
    def web3(self) -> Any:
        return self._web3

    @property
    def address(self) -> str:
        return self._address

    def to_wei(self, amount_ether: float) -> int:
        return self._web3.toWei(amount_ether, 'ether')

    def to_ether(self, amount_wei: int) -> float:
        return self._web3.fromWei(amount_wei, 'ether')

    def get_account_address(self, account_address: AccountAddress) -> str:
        address = account_address
        if hasattr(account_address, 'address'):
            address = account_address.address

        if not self.web3.isChecksumAddress(address):
            address = self.web3.toChecksumAddress(address)

        return address



"""

    Agent Manager


"""

from starfish.agent import RemoteAgent
from starfish.agent_manager.agent_access import AgentAccess
from starfish.network.ddo import DDO

LOCAL_AGENT_NAME = '_local_agent'


class AgentManager:

    def __init__(self, network=None):
        """
        Create an agent manager object to resolve agents and keep a list of knwon agents.
        If the list of known agents only have a url, then the agent manager will try
        to resolve the url's to DID's, using the api call to obtain the DDO from the
        remote agents.

        """
        self._agent_access_items = {}
        self._local_name = LOCAL_AGENT_NAME
        self._network = network

    def register_agents(self, agents, http_client=None):
        """
        Register a dict of agents and their url/did and access information.
        This allows the agent manager to resolve 'agent names'.
        So that you can pass a given name for the agent

        :param dict agents: Dict of agents that can be stored in the agent manager.
        :param object http_client: HTTP client to make the requests to resolve and load the agent

        The agents structure can have the following optional fields

        .. code-block:: yaml

            agent_name:
              url: URL of agent
              ddo: ddo object of agent
              ddo_text: ddo text of agent ( ddo.as_text )
              did: did of the agent
              authentication:
                username: username to acces the agent api
                password: password to acces the agent api
                token: token to access the agent api
              username: username to acces the agent api
              password: password to acces the agent api
              token: token to access the agent api

            # example config, both types of authentication will work

            surfer:
              url: http://localhost:3030
              username: Aladdin
              password: OpenSesame

            invokable:
              url: http://localhost:9090
              authentication:
                username: Aladdin
                password: OpenSesame


        """
        if not isinstance(agents, dict):
            raise TypeError('agents parameter must be a type dict')

        for name, agent in agents.items():
            self.register_agent(
                name,
                ddo_text=agent.get('ddo_text', None),
                ddo=agent.get('ddo', None),
                did=agent.get('did', None),
                url=agent.get('url', None),
                authentication=agent.get('authentication', None),
                username=agent.get('username', None),
                password=agent.get('password', None),
                token=agent.get('token', None),
                http_client=http_client
            )

    def register_agent(
        self,
        name,
        ddo_text=None,
        ddo=None,
        did=None,
        url=None,
        authentication=None,
        username=None,
        password=None,
        token=None,
        http_client=None
    ):
        """
        Register an agent with the agent manager object.

        :param str name: Name of the agent
        :param str ddo_text: DDO string of the local agent.
        :param DDO ddo: DDO object of the local agent.
        :param str did: Optional did of the agent.
        :param str url: Optional url of the agent, to resolve the ddo and did by calling the agent api.
        :param dict authentication: Authentication dict that can access the agent api
        :param str usrename: Username access for the agent api
        :param str password: Password to access the agent api
        :param str token: Token to access the agent api
        :param object http_client: Test http client to make requests to the agent

        """
        if ddo:
            ddo_text = ddo.as_text

        agent_access = AgentAccess(
            name,
            ddo_text=ddo_text,
            did=did,
            url=url,
            authentication=authentication,
            username=username,
            password=password,
            token=token,
            http_client=http_client
        )
        self._agent_access_items[name] = agent_access

    def unregister_agent(self, name_did_url):
        """

        Removes an agent from the access list of possible cached agents

        :returns: True if found and removed

        """
        for name, agent_access in self._agent_access_items.items():
            if agent_access.is_match(name_did_url):
                del self._agent_access_items[name]
                return True
        return False

    def register_local_agent(self, ddo_text, authentication=None, local_name=None, http_client=None):
        """
        Register a local agent with the  agent manager object.

        :param str ddo_text: DDO string of the local agent.
        :param dict authentication: Authentication dict that can access the local agent
        :param str local_name: Optional local name to use for this agent
        :param object http_client: Test http client to make requests to the agent

        """
        if isinstance(ddo_text, DDO):
            ddo_text = ddo_text.as_text

        if local_name:
            self._local_name = local_name
        agent_access = AgentAccess(self._local_name, ddo_text=ddo_text, authentication=authentication, http_client=http_client)
        self._agent_access_items[self._local_name] = agent_access

    def resolve_agent_url(self, url,  authentication=None, http_client=None):
        """
        Resolve an agent using ony the URL of the remote agent.

        :param str url: URL of the remote agent to resolve
        :param dict authentication: Authentication dict that can access the agent api
        :param object http_client: Test http client to make requests to the agent

        :returns: RemoteAgent object if found, else return None

        """
        ddo = AgentAccess.resolve_agent_url(url, authentication=authentication, http_client=http_client)
        if ddo:
            return RemoteAgent(ddo, authentication=authentication, http_client=http_client)

    def resolve_agent_did(self, did, network=None, authentication=None, http_client=None):
        """
        Resolve an agent using ony the DID of the remote agent. You need to set the network property
        on this object before searching for a did in the network

        :param str did: DID of the remote agent to resolve
        :param Network network: Optional network object to resolve the DID, if not used
            then use the class network value instead

        :param dict authentication: Authentication dict that can access the agent api
        :param object http_client: Test http client to make requests to the agent

        :returns: RemoteAgent object if found, else return None

        """
        if network is None:
            network = self._network
        if not network:
            raise ValueError('No network set to resolve a DID')
        ddo = AgentAccess.resolve_agent_did(did, network)
        if ddo:
            return RemoteAgent(ddo, authentication=authentication, http_client=http_client)

    def load_agent(self, name_did_url, authentication=None):
        """
        Loads and returns a starfish RemoteAgent that is found using the agent name/did/url.

        :param str name_did_url: name, did or url of the agent to find
        :param dict authentication: Optional authentication dict to access this server.

        :returns: RemoteAgent object for the loaded local agent.
        """

        found_agent_access = self.find_agent_access(name_did_url)
        if found_agent_access:
            return found_agent_access.load_agent(authentication=authentication)

    def load_ddo(self, name_did_url, authentication=None):
        """
        Loads and returns a starfish DDO that is found using the agent name/did/url.

        :param str name_did_url: name, did or url of the agent to find
        :param dict authentication: Optional authentication dict to access this server.

        :returns: RemoteAgent object for the loaded local agent.
        """

        found_agent_access = self.find_agent_access(name_did_url)
        if found_agent_access:
            return found_agent_access.ddo

    def find_agent_access(self, name_did_url, auto_resolve=True):
        """

        Finds an agent in the access records of this manager.
        If found return the agent access record

        :param str name_did_url: name/did or url of the agent to find in the access records
        :param bool auto_resolve: If True then if no agent found, try to resolve all of the agents,
            and call again with this field set too False.

        :returns: AgentAccess object if found
        """

        for name, agent_access in self._agent_access_items.items():
            if agent_access.is_match(name_did_url):
                # only return if we have a valid DDO
                if agent_access.ddo:
                    return agent_access

        if auto_resolve:
            self.resolve_access_agents()
            # call again but with no auto_resolve
            return self.find_agent_access(name_did_url, False)

    def is_agent(self, name_did_url):
        """
        Is the name/did or url in the agent access record list

        :param str name_did_url: name/did or url of the agent to find in the access records

        :returns: True if found

        """
        return self.find_agent_access(name_did_url) is not None

    def resolve_access_agents(self):
        """
        Resolve any agents that have only urls. This calls the agent api
        to get the DDO and DID from the agent

        """
        for name, agent_access in self._agent_access_items.items():
            if agent_access.ddo is None:
                if self._network and agent_access.did:
                    agent_access.resolve_did(self._network)
                elif agent_access.url:
                    agent_access.resolve_url()

    def set_http_client(self, value):
        """
        Set all of the agents with the same http_client value.
        This method also clears out any cache with the agent access items

        """

        for name, agent_item in self._agent_access_items.items():
            agent_item.http_client = value
            agent_item.clear_cache()

    def clear_cache(self):
        """
        Clears out the agent items cache.

        """
        for name, agent_item in self._agent_access_items.items():
            agent_item.clear_cache()

    @property
    def local_agent(self):
        """
        Returns the local agent, if it has been registered

        """
        return self.load_agent(self._local_name)

    @property
    def local_agent_name(self):
        return self._local_agent_name

    @property
    def items(self):
        return self._agent_access_items

    @property
    def network(self):
        return self._network

    @network.setter
    def network(self, value):
        self._network = value



"""

    Direct Purchase Contract

"""

from starfish.network.ethereum.ethereum_account import EthereumAccount

from .contract_base import ContractBase

CONTRACT_NAME = 'Dispenser'


class DispenserContract(ContractBase):
    """Class representing the Ocean token dispenser contract."""

    def __init__(self) -> None:
        ContractBase.__init__(self, CONTRACT_NAME)

    def request_tokens(self, account: EthereumAccount, amount: float) -> bool:
        # amount is in ether
        return self.call('requestTokens', amount, account)



"""
    Asset class to handle core imutable asset and it's metadata


"""

import json
from abc import ABC
from typing import (
    Any,
    Generic
)

from starfish.asset.provenance import Provenance
from starfish.types import TAssetBase
from starfish.utils.crypto_hash import hash_sha3_256


class AssetBase(ABC, Generic[TAssetBase]):
    """

    :param str metadata_text: metadata text for the asset
    :param did: Octional did of the asset, if the asset is new then the did will be None.
    :type did: None or str

    """
    def __init__(self, metadata_text: str) -> None:
        """
        init an asset class
        """
        if not isinstance(metadata_text, str):
            raise TypeError('metadata must be in text form')

        self._did = None
        self._metadata_text = metadata_text

        if 'name' not in self.metadata:
            raise ValueError('metadata must contain a metadata name')

        if 'type' not in self.metadata:
            raise ValueError('metadata must contain a metadata type')

        super().__init__()

    def add_provenance_publish(self, agent_did: str = None):
        """

        Add a published provenance data to the asset metadata.
        Calling this method will make the asset 'new'. So the asset_id will change,
        and the asset.did will be set to None.

        :param str agent_did: DID of the agent that this asset will be registered with

        """
        metadata = self.metadata
        provenance = Provenance(agent_did=agent_did)
        metadata['provenance'] = provenance.create_publish
        self.set_metadata(metadata)

    def add_provenance_invoke(self, agent_did: str, job_id: str, asset_list: Any, inputs_text: str):
        """

        Add a invoke provenance data to the asset metadata.
        Calling this method will make the asset 'new'. So the asset_id will change,
        and the asset.did will be set to None.

        :param str agent_did: DID of the agent that this asset will be registered with

        """
        metadata = self.metadata
        provenance = Provenance(agent_did=agent_did, activity_id=job_id, asset_list=asset_list, inputs_text=inputs_text)
        metadata['provenance'] = provenance.create_invoke
        self.set_metadata(metadata)

    def set_did(self, did: str) -> None:
        """
        This method makes the object immutable.
        So maybe a solution is that we have a 'copy' and
        set the did in the __init__ of the new class, to return a mutable copy of the
        same asset object.
        """
        self._did = did

    def is_asset_type(self, type_name: str) -> bool:
        """

        Returns if this metadata has the correct type

        :param str type_name: name of the asset type stored in the metadata

        :return: True if the metadata type is equal to type_name
        :type: boolean

        """
        asset_type = AssetBase.get_asset_type(self.metadata)
        return asset_type == type_name

    def set_metadata(self, metadata: any):
        self._metadata_text = json.dumps(metadata)
        self._did = None

    @property
    def did(self) -> str:
        """
        :return: the asset did
        :type: str
        """
        return self._did

    @property
    def metadata(self) -> Any:
        """

        WARNING: The read only version for this metadata, use `set_metadata` to assign a new metadata.
        Once you change metadata, the asset id is changed and the did is set to None.

        :return: The metadata for this asset
        :type: dict
        """
        return json.loads(self._metadata_text)

    @property
    def metadata_text(self) -> str:
        """
        :return: The metadata for this asset
        :type: dict
        """
        return self._metadata_text

    @property
    def asset_id(self) -> str:
        return hash_sha3_256(self._metadata_text)

    @property
    def name(self) -> str:
        return self.metadata['name']

    @property
    def type_name(self) -> str:
        return self.metadata['type']

    @property
    def is_bundle(self) -> bool:
        """

        Return True if this asset is a bundle asset and can contain sub assets ( Asset Bundle )

        :return: True if sub assets can be held within this asset
        :type: boolean

        """
        return False

    @property
    def is_new(self) -> bool:
        """

        Return True if this asset has not been saved, and can be changed.

        :return: True if no did has been set

        """
        return self._did is None

    @property
    def data(self) -> bytes:
        return None

    @staticmethod
    def get_asset_type(metadata: Any) -> str:
        asset_type = ''
        if isinstance(metadata, dict):
            if 'type' in metadata:
                asset_type = metadata['type']
            else:
                # if from squid then it's always a bundle
                if 'base' in metadata:
                    asset_type = 'bundle'
        return asset_type

    @staticmethod
    def generateMetadata(name: str, asset_type: str, metadata: Any = None) -> Any:
        if metadata is None:
            metadata = {}
        if not isinstance(metadata, dict):
            raise ValueError('The metadata has to be a dict')

        metadata['name'] = name
        metadata['type'] = asset_type
        return metadata



"""
    starfish.did contract

"""
import re

from convex_api import ConvexAPI
from convex_api.utils import (
    is_address,
    to_address
)

from starfish.network.convex.contract.contract_base import ContractBase
from starfish.network.convex.convex_account import ConvexAccount
from starfish.network.did import (
    did_to_id,
    id_to_did
)
from starfish.types import AccountAddress


class DIDContract(ContractBase):

    def __init__(self, convex: ConvexAPI):
        ContractBase.__init__(self, convex, 'starfish.did')

    def register_did(self, did: str, ddo_text: str, account: ConvexAccount):
        encode_ddo_text = re.sub('\\\\', '\\\\\\\\', ddo_text)
        encode_ddo_text = re.sub('"', '\\"', encode_ddo_text)
        did_id = did_to_id(did).lower()
        command = f'(register {did_id} "{encode_ddo_text}")'
        result = self.send(command, account)
        if result and 'value' in result:
            return id_to_did(result['value'])
        return result

    def resolve(self, did: str, account_address: AccountAddress):
        did_id = did_to_id(did).lower()
        command = f'(resolve {did_id})'
        if is_address(account_address):
            address = account_address
        else:
            address = account_address.address
        result = self.query(command, address)
        if result and 'value' in result:
            return result['value']
        return result

    def owner(self, did: str,  account_address: AccountAddress):
        did_id = did_to_id(did).lower()
        command = f'(owner {did_id})'
        if is_address(account_address):
            address = account_address
        else:
            address = account_address.address
        result = self.query(command, address)
        if result and 'value' in result:
            return to_address(result['value'])
        return result



"""

    DDO for starfish


"""
import json
import re

from typing import Any

from starfish.network.did import did_generate_random
from starfish.utils.crypto_hash import hash_sha3_256


class DDO:

    SUPPORTED_SERVICES = {
        'meta': {
            'type': 'DEP.Meta',
            'uri': '/meta',
        },
        'storage': {
            'type': 'DEP.Storage',
            'uri': '/assets',
        },
        'invoke': {
            'type': 'DEP.Invoke',
            'uri': '/invoke',
        },
        'market': {
            'type': 'DEP.Market',
            'uri': '/market',
        },
        'trust': {
            'type': 'DEP.Trust',
            'uri': '/trust',
        },
        'auth': {
            'type': 'DEP.Auth',
            'uri': '/auth',
        },
        'collection': {
            'type': 'DEP.Collection',
            'uri': '/collection',
        }
    }

    DEFAULT_VERSION = 'v1'

    @staticmethod
    def create(url, service_list=None, version=None, did=None):
        if service_list:
            if not isinstance(service_list, (tuple, list)):
                raise TypeError('Service list must be a list or tuple of service names')
            elif isinstance(service_list, dict):
                # serice_list is a dict of values
                return DDO(did, service_list)
        else:
            service_list = list(DDO.SUPPORTED_SERVICES.keys())

        ddo = DDO(did)
        for name in service_list:
            if not isinstance(name, str):
                raise TypeError('Service list must be a list of type string names')
            ddo.add_service(name, url=url, version=version)
        return ddo

    @staticmethod
    def import_from_text(text):
        data = json.loads(text)
        return DDO(data['id'], data['service'])

    @staticmethod
    def is_supported_service(name):
        return name in DDO.SUPPORTED_SERVICES

    @staticmethod
    def get_did_from_ddo(ddo_data: Any) -> str:
        if isinstance(ddo_data, str):
            ddo = DDO.import_from_text(ddo_data)
        else:
            ddo = ddo_data
        if ddo:
            return ddo.did

    def __init__(self, did=None, service=None):
        if did is None:
            did = did_generate_random()
        self._did = did
        self._service = {}
        if service:
            if not isinstance(service, (tuple, list)):
                raise TypeError('Service data must be a tuple or list')
            for item in service:
                service_type = item['type']
                name = DDO._supported_service_name_from_type(service_type)
                if not name:
                    name = service_type.lower()
                self._service[name] = item

    def add_service(self, name, endpoint=None, url=None, version=None):
        if not version:
            version = DDO.DEFAULT_VERSION
        if re.search(r'\.', name):
            service_type = name
            # now try to resolve name based on the service type
            name = DDO._supported_service_name_from_type(service_type)
            if not name:
                # if cannot resolve thet set to the new service_type ( orignal name )
                name = service_type

        else:
            # using a template to setup the service
            if name not in DDO.SUPPORTED_SERVICES:
                raise ValueError(f'Invalid service name {name}')
            template = DDO.SUPPORTED_SERVICES[name]
            service_type = f'{template["type"]}.{version}'
            if url:
                endpoint = f'{url}/api/{version}{template["uri"]}'

        if not endpoint:
            raise ValueError('No endpoint or url provided to add a service to the ddo')

        service = {
            'type': service_type,
            'serviceEndpoint': endpoint
        }
        self._service[name] = service

    def remove_service(self, name):
        if name in self._service:
            del self._service[name]
            return True
        return False

    def is_sevice(self, name):
        if name in self._service:
            return True
        return False

    def get_service(self, name):
        if name in self._service:
            return self._service[name]

    def get_service_type(self, service_type):
        for name, item in self._service.items():
            if item['type'] == service_type:
                return item

    @property
    def service_list(self):
        items = []
        for name in sorted(self._service):
            items.append(self._service[name])
        return items

    @property
    def service(self):
        return self._service

    @property
    def did(self):
        return self._did

    @property
    def as_text(self):
        return DDO._as_json(did=self._did, service_list=self.service_list)

    @property
    def checksum(self):
        return hash_sha3_256(DDO._as_json(service_list=self.service_list))

    @staticmethod
    def _as_json(did=None, service_list=None):
        values = {
            '@context': 'https://www.w3.org/2019/did/v1',
        }
        if did:
            values['id'] = did
        if service_list:
            values['service'] = service_list
        return json.dumps(values, sort_keys=True)

    @staticmethod
    def _supported_service_name_from_type(service_type):
        match = re.match(r'^DEP\.(\w+)\.', service_type)
        if match:
            service_name = match.group(1)
            regexp = fr'DEP\.{service_name}'
            for name, item in DDO.SUPPORTED_SERVICES.items():
                if re.match(regexp, item['type']):
                    return name



"""

    Command 'Agent Resolve'

"""
from typing import Any

from .command_base import CommandBase


class AgentResolveCommand(CommandBase):

    def __init__(self, sub_parser: Any = None) -> None:
        self._command_list = []
        super().__init__('resolve', sub_parser)

    def create_parser(self, sub_parser: Any) -> Any:

        parser = sub_parser.add_parser(
            self._name,
            description='Tool to resolve an agent',
            help='Tool to resolve an agent using its DID or URL',

        )

        parser.add_argument(
            '-u',
            '--username',
            help='Username to access the remote agent'
        )

        parser.add_argument(
            '-p',
            '--password',
            help='Password to access the remote agent'
        )

        parser.add_argument(
            'agent',
            help='agent url or agent did to resolve. If url, you may need to provide --username and --password options'
        )

        return parser

    def execute(self, args: Any, output: Any) -> None:
        network = self.get_network(args.url)

        output.set_value('agent', args.agent)

        ddo = network.resolve_agent(args.agent, username=args.username, password=args.password)
        if ddo:
            output.add_line(f'{args.agent} resolved to {ddo.as_text}')
            output.add_line(f'did: {ddo.did}')
            output.set_value('ddo_text', ddo.as_text)
            output.set_value('did', ddo.did)
        else:
            output.add_line(f'unable to resolve agent address {args.agent}')



"""


    Starfish Tool Output

"""
import json
from typing import Any


class Output:
    def __init__(self) -> None:
        self._line_list = []
        self._values = {}
        self._error_list = []

    def add_error(self, line: str) -> None:
        self._error_list.append(line)

    def add_line(self, line: str) -> None:
        self._line_list.append(line)

    def set_value(self, key: str, value: str) -> None:
        self._values[key] = value

    def printout(self, is_json: bool) -> None:
        if self.has_errors:
            error_lines = '\nError: '. join(self.errors)
            print(f'Error: {error_lines}')
            return

        if is_json:
            if self.has_values:
                print(json.dumps(self.values, sort_keys=True, indent=2))
        else:
            if self.has_lines:
                print('\n'.join(self.lines))

    @property
    def errors(self) -> Any:
        return self._error_list

    @property
    def has_errors(self) -> bool:
        return len(self._error_list) > 0

    @property
    def lines(self) -> Any:
        return self._line_list

    @property
    def has_lines(self) -> bool:
        return len(self._line_list) > 0

    @property
    def values(self) -> Any:
        return self._values

    @property
    def has_values(self) -> bool:
        return len(self._values.keys()) > 0



"""

    Command Account Send ..

"""
from typing import Any

from .account_send_ether_command import AccountSendEtherCommand
from .account_send_token_command import AccountSendTokenCommand
from .command_base import CommandBase
from .help_command import HelpCommand

DEFAULT_AMOUNT = 10


class AccountSendCommand(CommandBase):

    def __init__(self, sub_parser: Any = None) -> None:
        self._command_list = []
        super().__init__('send', sub_parser)

    def create_parser(self, sub_parser: Any) -> Any:

        parser = sub_parser.add_parser(
            self._name,
            description='Send ether or tokens from one account to another',
            help='Send some tokens or ether from one account to another'
        )

        account_send_parser = parser.add_subparsers(
            title='send command',
            description='send command',
            help='Send command',
            dest='send_command'
        )

        self._command_list = [
            AccountSendEtherCommand(account_send_parser),
            AccountSendTokenCommand(account_send_parser),
            HelpCommand(account_send_parser, self)
        ]
        return account_send_parser

    def execute(self, args: Any, output: Any) -> Any:
        return self.process_sub_command(args, output, args.send_command)
